/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Peers facility.
 *
 * CER / CEA messages management
 *
 * See peer-api.h and peer.h for more information on the functions 
 *
 */

#include "waaad-internal.h"
#include "peer-internal.h"
#include "hgversion.h"

/* Dictionary objects for PSM */
static dict_object_t * dict_CER = NULL;
static dict_object_t * dict_CEA = NULL;

int _peer_cer_cea_init(void)
{
	CHECK_FCT( dict_search( DICT_COMMAND, CMD_BY_NAME, "Capabilities-Exchange-Request", &dict_CER) );
	CHECK_FCT( dict_search( DICT_COMMAND, CMD_BY_NAME, "Capabilities-Exchange-Answer",  &dict_CEA) );
	
	return 0;
}

int _peer_cer_cea_fini(void)
{
	/* nothing special to do */
	return 0;
}

/******************************************************************************/
/*  Diameter Base Protocol: Capabilities-Exchanges messages                   */
/******************************************************************************/
/* Create a CER message */
int _peer_cer_create( _peer_t * peer, msg_t ** cer )
{
	dict_object_t * dict = NULL;
	msg_avp_t * avp = NULL;
	avp_value_t val;
	int i;
	
	TRACE_ENTRY("%p", cer);
	
	/* NULL is not allowed */
	CHECK_PARAMS( cer );
	
	/* Create an empty CER */
	CHECK_FCT( msg_new( dict_CER, 0, cer ) );
	
	/* Add standard Origin-Host, Origin-Realm and Origin-State-Id AVPs */
	CHECK_FCT( msg_add_origin( *cer, 1 ) );
	
	/* Add the Host-IP-Address AVP(s) */
	{
		/* Find the model for this AVP */
		CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Host-IP-Address", &dict )  );
		
		switch (peer->p_addinfo.pa_proto) {
			case IPPROTO_TCP: /* In TCP, we have only one endpoint. We must add only one Host-IP-Address AVP */
			{
				/* Search for an address in local_addr_pri_sa */
				sSS * ss;
				int found = 0;

				for (ss = g_pconf->local_addr_pri_sa; ss && ss->ss_family != 0; ss++) {
					if (ss->ss_family == peer->p_addinfo.pa_ss.ss_family) {
						found = 1;
						break;
					}
				}

				if (!found) {
					if (peer->p_addinfo.pa_ss.ss_family == AF_INET)
						ss = &g_pconf->local_addr_diamid[0];
					else
						ss = &g_pconf->local_addr_diamid[1];
				}

				/* ss should never be NULL at this point */
				CHECK_PARAMS( ss );

				/* Now create the AVP instance */
				CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
				
				/* Set its value */
				CHECK_FCT( msg_avp_value_encode( ss, avp ) );
				
				/* Add to the message */
				CHECK_FCT( msg_avp_add( *cer, MSG_BRW_LAST_CHILD, avp ) );
			}
			break;
			
			case IPPROTO_SCTP: /* In SCTP, we send all the declared endpoints... will probably need change! */
			{
				sSS * ss;
				int done = 0;
				
				/* Add all the elements from primary and secondary list */
				for (ss = g_pconf->local_addr_pri_sa; ss && ss->ss_family != 0; ss++) {
					CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
					CHECK_FCT( msg_avp_value_encode( ss, avp ) );
					CHECK_FCT( msg_avp_add( *cer, MSG_BRW_LAST_CHILD, avp ) );
					done ++;
				}
				for (ss = g_pconf->local_addr_sec_sa; ss && ss->ss_family != 0; ss++) {
					CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
					CHECK_FCT( msg_avp_value_encode( ss, avp ) );
					CHECK_FCT( msg_avp_add( *cer, MSG_BRW_LAST_CHILD, avp ) );
					done ++;
				}
				
				if (!done) {
					if (g_pconf->local_addr_diamid[0].ss_family != 0) {
						CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
						CHECK_FCT( msg_avp_value_encode( &g_pconf->local_addr_diamid[0], avp ) );
						CHECK_FCT( msg_avp_add( *cer, MSG_BRW_LAST_CHILD, avp ) );
					}
					if (g_pconf->local_addr_diamid[1].ss_family != 0) {
						CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
						CHECK_FCT( msg_avp_value_encode( &g_pconf->local_addr_diamid[1], avp ) );
						CHECK_FCT( msg_avp_add( *cer, MSG_BRW_LAST_CHILD, avp ) );
					}
				}
			}
			break;
			
			default: /* ??? */
				TRACE_DEBUG(INFO, "Unsupported proto for peer %s", peer->p_diamid);
				ASSERT( 0 ); /* Unsupported protocol */
				
		}
	}
	
	/* Add the Vendor-Id, Product-Name, and Firmware-Revision AVPs */
	{
		/* Find the model for this AVP */
		CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Vendor-Id", &dict )  );
		
		/* Create the AVP */
		CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
		
		/* Set its value */
		memset(&val, 0, sizeof(val));
		val.u32 = MY_VENDOR_ID;
		CHECK_FCT( msg_avp_setvalue( avp, &val ) );

		/* Add it to the message */
		CHECK_FCT( msg_avp_add( *cer, MSG_BRW_LAST_CHILD, avp ) );
	}
	{
		/* Find the model for this AVP */
		CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Product-Name", &dict )  );
		
		/* Create the AVP */
		CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
		
		/* Set its value */
		memset(&val, 0, sizeof(val));
		val.os.data = (unsigned char *)PACKAGE_TARNAME;
		val.os.len = strlen(PACKAGE_TARNAME);
		CHECK_FCT( msg_avp_setvalue( avp, &val ) );

		/* Add it to the message */
		CHECK_FCT( msg_avp_add( *cer, MSG_BRW_LAST_CHILD, avp ) );
	}
	{
		/* Find the model for this AVP */
		CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Firmware-Revision", &dict )  );
		
		/* Create the AVP */
		CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
		
		/* Set its value */
		memset(&val, 0, sizeof(val));
		#ifdef HG_REVISION
		val.u32 = (uint32_t)atol(HG_REVISION);
		#else /* HG_REVISION */
		val.u32 = (uint32_t)atol(PACKAGE_VERSION); /* Will convert only the major of the version... */
		#endif /* HG_REVISION */

		CHECK_FCT( msg_avp_setvalue( avp, &val ) );

		/* Add it to the message */
		CHECK_FCT( msg_avp_add( *cer, MSG_BRW_LAST_CHILD, avp ) );
	}
		
	/* Add the Inband-Security-Id AVP */
	{
		ASSERT( ! IS_LIST_EMPTY(&peer->p_sec_list)  );
		
		/* Find the model for this AVP */
		CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Inband-Security-Id", &dict )  );
		
		/* Create the AVP */
		CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
		
		/* Set its value */
		memset(&val, 0, sizeof(val));
		val.u32 = ((sec_module_t *)(peer->p_sec_list.next->o))->sec_insecid;
		CHECK_FCT( msg_avp_setvalue( avp, &val ) );

		/* Add it to the message */
		CHECK_FCT( msg_avp_add( *cer, MSG_BRW_LAST_CHILD, avp ) );
	}
	
	/* Add the list of supported applications and vendors */
	{
		/* Supported-Vendor-Id: all vendors from the dictionary, except MY_VENDOR_ID */
		{
			int i = 0;
			uint32_t *array = NULL;
			
			array = dict_get_vendorid_list();
			
			if ((array != NULL) && (array[0] != 0)) {
				/* Find the model for this AVP */
				CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Supported-Vendor-Id", &dict )  );

				memset(&val, 0, sizeof(val));
				
				while (array[i] != 0) {
					/* Create the AVP */
					CHECK_FCT( msg_avp_new( dict, 0, &avp ) );

					/* Set its value */
					val.u32 = array[i];
					CHECK_FCT( msg_avp_setvalue( avp, &val ) );

					/* Add it to the message */
					CHECK_FCT( msg_avp_add( *cer, MSG_BRW_LAST_CHILD, avp ) );
					
					/* go to next vendor */
					i++;
				}
				
			}
			if (array)
				free(array);
		}
		
		/* Auth-Application-Id, Acct-Application-Id, Vendor-Specific-Application-Id: set in the configuration by the dispatch module */
		for ( i = 0; i < g_conf->supported_apps_nb; i++) {
			
			memset(&val, 0, sizeof(val));
			
			/* Create the application AVP */
			if (g_conf->supported_apps_list[i].t == APP_TYPE_AUTH) {
				
				CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Auth-Application-Id", &dict )  );

			} else {
				
				CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Acct-Application-Id", &dict )  );

			}
			
			/* Create the AVP */
			CHECK_FCT( msg_avp_new( dict, 0, &avp ) );

			/* Set its value */
			val.u32 = g_conf->supported_apps_list[i].a;

			CHECK_FCT( msg_avp_setvalue( avp, &val ) );
			
			/* In case of Vendor-specific, create the parent AVP */
			if (g_conf->supported_apps_list[i].v != 0)  {
				msg_avp_t * avp_v = NULL;
				msg_avp_t * avp_a = avp;
				
				/* The Vendor-Id AVP */
				CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Vendor-Id", &dict )  );
				CHECK_FCT( msg_avp_new( dict, 0, &avp_v ) );
				memset(&val, 0, sizeof(val));
				val.u32 = g_conf->supported_apps_list[i].v;
				CHECK_FCT( msg_avp_setvalue( avp_v, &val ) );
				
				/* The grouped AVP */
				CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Vendor-Specific-Application-Id", &dict )  );
				CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
				
				/* Add the children AVPs */
				CHECK_FCT( msg_avp_add( avp, MSG_BRW_LAST_CHILD, avp_v ) );
				CHECK_FCT( msg_avp_add( avp, MSG_BRW_LAST_CHILD, avp_a ) );
			}
			
			/* Now add this AVP to the message */
			CHECK_FCT( msg_avp_add( *cer, MSG_BRW_LAST_CHILD, avp ) );
		}
		
		/* The Relay application */
		if (g_pconf->disable_relay == 0) {
			dict_object_t * apprelay = NULL;
			dict_application_data_t relay;
			
			/* Find the Relay application in the dictionary */
			CHECK_FCT(  dict_search( DICT_APPLICATION, APPLICATION_BY_NAME, "Relay", &apprelay )  );
			CHECK_FCT(  dict_getval( apprelay, &relay )  );
			
			/* Find the model for this AVP */
			CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Auth-Application-Id", &dict )  );

			/* Create the AVP */
			CHECK_FCT( msg_avp_new( dict, 0, &avp ) );

			/* Set its value */
			memset(&val, 0, sizeof(val));
			val.u32 = relay.application_id;
			CHECK_FCT( msg_avp_setvalue( avp, &val ) );

			/* Add it to the message */
			CHECK_FCT( msg_avp_add( *cer, MSG_BRW_LAST_CHILD, avp ) );
		}
	}
	
	/* Set the CER header: end-to-end id */
	{
		msg_data_t * cer_data = NULL;
		
		CHECK_FCT(  msg_data( *cer, &cer_data )  );
		
		cer_data->msg_eteid = msg_get_eteid();
	}
	
	/* Make sure the CER message is valid -- mostly for debug here... */
	CHECK_FCT(  msg_parse_rules( *cer, NULL )  );
	
	/* The CER is now ready */
	TRACE_DEBUG(FULL, "CER Ready");
	
	return 0;
}

/* Stores the values of the CER AVPs in the peer structure */
int _peer_cer_parse( msg_t * cer, _peer_t * peer )
{
	int i;
	msg_avp_t * avp = NULL;
	
	CHECK_PARAMS(  VALIDATE_PEER( peer ) && cer  );

	CHECK_FCT(  msg_browse(cer, MSG_BRW_FIRST_CHILD, &avp, NULL)  );

	/* Now loop on all AVPs -- we will break when last AVP was parsed */
	while (avp) {
		msg_avp_data_t * avpdata;

		CHECK_FCT(  msg_avp_data( avp, &avpdata )  );

		if (avpdata->avp_flags & AVP_FLAG_VENDOR) {
			/* Ignore all vendor-specific AVPs in CER ... because we don't support any currently */
			TRACE_DEBUG(FULL, "Ignored a vendor AVP in CER");
			msg_dump_one(FULL, avp);
			goto next;
		}

		switch (avpdata->avp_code) {
			case AC_ORIGIN_HOST: /* Origin-Host */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CER");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}

				/* Copy the value into peer->p_diamid, as a lowercase string */
				free(peer->p_diamid);
				CHECK_MALLOC(  peer->p_diamid = (char *) malloc( avpdata->avp_data->os.len + 1 )  );
				for (i=0; i<avpdata->avp_data->os.len; i++) {
					peer->p_diamid[i] = (char)tolower((char)avpdata->avp_data->os.data[i]);
				}
				peer->p_diamid[i] = '\0';
				break;

			case AC_ORIGIN_REALM: /* Origin-Realm */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CER");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}
				/* Copy the value into peer->p_realm, as a lowercase string */
				free(peer->p_realm);
				CHECK_MALLOC(  peer->p_realm = (char *) malloc( avpdata->avp_data->os.len + 1 )  );
				for (i=0; i<avpdata->avp_data->os.len; i++) {
					peer->p_realm[i] = (char)tolower((char)avpdata->avp_data->os.data[i]);
				}
				peer->p_realm[i] = '\0';
				break;

			case AC_HOST_IP_ADDRESS: /* Host-IP-Address */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CER");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}
				if (peer->p_peeraddr_sz == 0) {
					CHECK_MALLOC(  peer->p_peeraddr = (sSS *) malloc( sizeof(sSS) ) );
				} else {
					CHECK_MALLOC(  peer->p_peeraddr = (sSS *) realloc( peer->p_peeraddr, 
										(peer->p_peeraddr_sz + 1 ) * sizeof(sSS) )  );
				}

				/* Add this AVP as last element */
				CHECK_FCT( msg_avp_value_interpret( avp, &(peer->p_peeraddr[peer->p_peeraddr_sz]) ) );

				/* Update the number of elements in the array */
				peer->p_peeraddr_sz += 1;

				break;


			case AC_VENDOR_ID: /* Vendor-Id */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CER");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}
				peer->p_vendor = avpdata->avp_data->u32;
				break;

			case AC_PRODUCT_NAME: /* Product-Name */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CER");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}
				free(peer->p_prodname);

				CHECK_MALLOC(  peer->p_prodname = (char *) malloc( avpdata->avp_data->os.len + 1 )  );
				memcpy(peer->p_prodname, avpdata->avp_data->os.data, avpdata->avp_data->os.len);
				peer->p_prodname[avpdata->avp_data->os.len] = '\0';
				break;

			case AC_ORIGIN_STATE_ID: /* Origin-State-Id */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CER");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}
				peer->p_orstate = avpdata->avp_data->u32;
				break;

			case AC_SUPPORTED_VENDOR_ID: /* Supported-Vendor-Id */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CER");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}
				TRACE_DEBUG(FULL, "Info: The remote peer claims support for (a subset of) vendor %d features...", avpdata->avp_data->u32);
				break;

			case AC_VENDOR_SPECIFIC_APPLICATION_ID: /* Vendor-Specific-Application-Id (grouped)*/
				{
					peer_appl_t appdata;
					msg_avp_t * inavp = NULL;

					memset(&appdata, 0, sizeof(appdata));

					/* get the first child AVP */
					CHECK_FCT(  msg_browse(avp, MSG_BRW_FIRST_CHILD, &inavp, NULL)  );

					while (inavp) {
						msg_avp_data_t * inavpdata;
						CHECK_FCT(  msg_avp_data( inavp, &inavpdata )  );

						if (inavpdata->avp_flags & AVP_FLAG_VENDOR) {
							/* Ignore all vendor-specific AVPs in CER ... because we don't support any currently */
							TRACE_DEBUG(FULL, "Ignored a vendor AVP in Vendor-Specific-Application-Id AVP");
							msg_dump_one(FULL, avp);
							goto innext;
						}

						if (inavpdata->avp_data == NULL) {
							/* Ignore if the data is not set -- this is more like a sanity check */
							TRACE_DEBUG(FULL, "Ignored an AVP with unset value in Vendor-Specific-Application-Id AVP");
							msg_dump_one(FULL, avp);
							ASSERT(0); /* To check if this really happens, and understand why... */
							goto innext;
						}
						switch (inavpdata->avp_code) {
							case AC_VENDOR_ID: /* Vendor-Id */
								appdata.v = inavpdata->avp_data->u32;
								break;
							case AC_AUTH_APPLICATION_ID: /* Auth-Application-Id */
								appdata.a = inavpdata->avp_data->u32;
								appdata.t = APP_TYPE_AUTH;
								break;
							case AC_ACCT_APPLICATION_ID: /* Acct-Application-Id */
								appdata.a = inavpdata->avp_data->u32;
								appdata.t = APP_TYPE_ACCT;
								break;
							default: /* Other AVP */
								TRACE_DEBUG(FULL, "Ignored AVP in Vendor-Specific-Application-Id AVP");
								msg_dump_one(FULL, inavp);
						}

					innext:			
						/* Go to next in AVP */
						CHECK_FCT( msg_browse(inavp, MSG_BRW_NEXT, &inavp, NULL) );
					}

					CHECK_FCT(  _peer_struct_appl_add(peer, appdata.a, appdata.v, appdata.t)  );
				}
				break;

			case AC_AUTH_APPLICATION_ID: /* Auth-Application-Id */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CER");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}
				CHECK_FCT(  _peer_struct_appl_add(peer, avpdata->avp_data->u32, 0, APP_TYPE_AUTH)  );
				break;

			case AC_ACCT_APPLICATION_ID: /* Acct-Application-Id */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CER");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}
				CHECK_FCT(  _peer_struct_appl_add(peer, avpdata->avp_data->u32, 0, APP_TYPE_ACCT)  );
				break;

			case AC_INBAND_SECURITY_ID: /* Inband-Security-Id */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CER");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}
				CHECK_FCT(  _peer_struct_secid_add(peer, avpdata->avp_data->u32)  );
				break;

			case AC_FIRMWARE_REVISION: /* Firmware-Revision */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CER");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}
				peer->p_firmrev = avpdata->avp_data->u32;
				break;

			default: /* Other AVP */
				TRACE_DEBUG(FULL, "Ignored AVP in CER");
				msg_dump_one(FULL, avp);
		}

next:			
		/* Go to next AVP */
		CHECK_FCT( msg_browse(avp, MSG_BRW_NEXT, &avp, NULL) );
	}
	
	return 0;
}

int _peer_cea_create( _peer_t * peer, msg_t ** cea, uint32_t eteid, uint32_t hbhid )
{
	dict_object_t * dict = NULL;
	msg_avp_t * avp = NULL;
	avp_value_t val;
	int i;
	
	TRACE_ENTRY("%p %p %u %u", peer, cea, eteid, hbhid);
	
	/* NULL is not allowed */
	CHECK_PARAMS( cea );
	
	/* Create an empty CEA */
	CHECK_FCT( msg_new( dict_CEA, 0, cea ) );
	
	/* Note: the Result-Code AVP will be added later outside this function */
	
	/* Add standard Origin-Host, Origin-Realm and Origin-State-Id AVPs */
	CHECK_FCT( msg_add_origin( *cea, 1 ) );
	
	/* Add the Host-IP-Address AVP(s) */
	{
		/* Find the model for this AVP */
		CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Host-IP-Address", &dict )  );
		
		switch (peer->p_addinfo.pa_proto) {
			case IPPROTO_TCP: /* In TCP, we have only one endpoint. We must add only one Host-IP-Address AVP */
			{
				/* Since we are connected already, just read the socket name */
				sSS ss;
				socklen_t len = sizeof(sSS);
				
				CHECK_SYS( getsockname(peer->p_sock, (sSA *)&ss, &len) );
				
				/* Now create the AVP instance */
				CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
				
				/* Set its value */
				CHECK_FCT( msg_avp_value_encode( &ss, avp ) );
				
				/* Add to the message */
				CHECK_FCT( msg_avp_add( *cea, MSG_BRW_LAST_CHILD, avp ) );
			}
			break;
			
			case IPPROTO_SCTP: /* In SCTP, we send all the declared endpoints... will probably need change! */
			{
				sSS * ss;
				int done = 0;
				
				/* Add all the elements from primary and secondary list */
				for (ss = g_pconf->local_addr_pri_sa; ss && ss->ss_family != 0; ss++) {
					CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
					CHECK_FCT( msg_avp_value_encode( ss, avp ) );
					CHECK_FCT( msg_avp_add( *cea, MSG_BRW_LAST_CHILD, avp ) );
					done ++;
				}
				for (ss = g_pconf->local_addr_sec_sa; ss && ss->ss_family != 0; ss++) {
					CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
					CHECK_FCT( msg_avp_value_encode( ss, avp ) );
					CHECK_FCT( msg_avp_add( *cea, MSG_BRW_LAST_CHILD, avp ) );
					done ++;
				}
				
				if (!done) {
					if (g_pconf->local_addr_diamid[0].ss_family != 0) {
						CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
						CHECK_FCT( msg_avp_value_encode( &g_pconf->local_addr_diamid[0], avp ) );
						CHECK_FCT( msg_avp_add( *cea, MSG_BRW_LAST_CHILD, avp ) );
					}
					if (g_pconf->local_addr_diamid[1].ss_family != 0) {
						CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
						CHECK_FCT( msg_avp_value_encode( &g_pconf->local_addr_diamid[1], avp ) );
						CHECK_FCT( msg_avp_add( *cea, MSG_BRW_LAST_CHILD, avp ) );
					}
				}
			}
			break;
			
			default: /* ??? */
				TRACE_DEBUG(INFO, "Unsupported proto for peer %s", peer->p_diamid);
				ASSERT( 0 ); /* Unsupported protocol */
				
		}
	}
	
	/* Add the Vendor-Id, Product-Name, and Firmware-Revision AVPs */
	{
		/* Find the model for this AVP */
		CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Vendor-Id", &dict )  );
		
		/* Create the AVP */
		CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
		
		/* Set its value */
		memset(&val, 0, sizeof(val));
		val.u32 = MY_VENDOR_ID;
		CHECK_FCT( msg_avp_setvalue( avp, &val ) );

		/* Add it to the message */
		CHECK_FCT( msg_avp_add( *cea, MSG_BRW_LAST_CHILD, avp ) );
	}
	{
		/* Find the model for this AVP */
		CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Product-Name", &dict )  );
		
		/* Create the AVP */
		CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
		
		/* Set its value */
		memset(&val, 0, sizeof(val));
		val.os.data = (unsigned char *)PACKAGE_TARNAME;
		val.os.len = strlen(PACKAGE_TARNAME);
		CHECK_FCT( msg_avp_setvalue( avp, &val ) );

		/* Add it to the message */
		CHECK_FCT( msg_avp_add( *cea, MSG_BRW_LAST_CHILD, avp ) );
	}
	{
		/* Find the model for this AVP */
		CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Firmware-Revision", &dict )  );
		
		/* Create the AVP */
		CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
		
		/* Set its value */
		memset(&val, 0, sizeof(val));
		#ifdef HG_REVISION
		val.u32 = (uint32_t)atol(HG_REVISION);
		#else /* HG_REVISION */
		val.u32 = (uint32_t)atol(PACKAGE_VERSION); /* Will convert only the major of the version... */
		#endif /* HG_REVISION */

		CHECK_FCT( msg_avp_setvalue( avp, &val ) );

		/* Add it to the message */
		CHECK_FCT( msg_avp_add( *cea, MSG_BRW_LAST_CHILD, avp ) );
	}
		
	/* Add the Inband-Security-Id AVP, if found */
	if (peer->p_sec_hdl != NULL)
	{
		sec_module_t * mod = NULL;
		
		/* Retrieve the module */
		CHECK_FCT(  sec_getmodfromhdl(peer->p_sec_hdl, &mod)  );
		
		/* Find the model for this AVP */
		CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Inband-Security-Id", &dict )  );
		
		/* Create the AVP */
		CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
		
		/* Set its value */
		memset(&val, 0, sizeof(val));
		val.u32 = mod->sec_insecid;
		CHECK_FCT( msg_avp_setvalue( avp, &val ) );

		/* Add it to the message */
		CHECK_FCT( msg_avp_add( *cea, MSG_BRW_LAST_CHILD, avp ) );
	}
	
	/* Add the list of supported applications and vendors */
	{
		/* Supported-Vendor-Id: all vendors from the dictionary, except MY_VENDOR_ID */
		{
			int i = 0;
			uint32_t *array = NULL;
			
			array = dict_get_vendorid_list();
			
			if ((array != NULL) && (array[0] != 0)) {
				/* Find the model for this AVP */
				CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Supported-Vendor-Id", &dict )  );

				memset(&val, 0, sizeof(val));
				
				while (array[i] != 0) {
					/* Create the AVP */
					CHECK_FCT( msg_avp_new( dict, 0, &avp ) );

					/* Set its value */
					val.u32 = array[i];
					CHECK_FCT( msg_avp_setvalue( avp, &val ) );

					/* Add it to the message */
					CHECK_FCT( msg_avp_add( *cea, MSG_BRW_LAST_CHILD, avp ) );
					
					/* go to next vendor */
					i++;
				}
				
			}
			if (array)
				free(array);
		}
		
		/* Auth-Application-Id, Acct-Application-Id, Vendor-Specific-Application-Id: set in the configuration by the dispatch module */
		for ( i = 0; i < g_conf->supported_apps_nb; i++) {
			
			memset(&val, 0, sizeof(val));
			
			/* Create the application AVP */
			if (g_conf->supported_apps_list[i].t == APP_TYPE_AUTH) {
				
				CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Auth-Application-Id", &dict )  );

			} else {
				
				CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Acct-Application-Id", &dict )  );

			}
			
			/* Create the AVP */
			CHECK_FCT( msg_avp_new( dict, 0, &avp ) );

			/* Set its value */
			val.u32 = g_conf->supported_apps_list[i].a;

			CHECK_FCT( msg_avp_setvalue( avp, &val ) );
			
			/* In case of Vendor-specific, create the parent AVP */
			if (g_conf->supported_apps_list[i].v != 0)  {
				msg_avp_t * avp_v = NULL;
				msg_avp_t * avp_a = avp;
				
				/* The Vendor-Id AVP */
				CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Vendor-Id", &dict )  );
				CHECK_FCT( msg_avp_new( dict, 0, &avp_v ) );
				memset(&val, 0, sizeof(val));
				val.u32 = g_conf->supported_apps_list[i].v;
				CHECK_FCT( msg_avp_setvalue( avp_v, &val ) );
				
				/* The grouped AVP */
				CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Vendor-Specific-Application-Id", &dict )  );
				CHECK_FCT( msg_avp_new( dict, 0, &avp ) );
				
				/* Add the children AVPs */
				CHECK_FCT( msg_avp_add( avp, MSG_BRW_LAST_CHILD, avp_v ) );
				CHECK_FCT( msg_avp_add( avp, MSG_BRW_LAST_CHILD, avp_a ) );
			}
			
			/* Now add this AVP to the message */
			CHECK_FCT( msg_avp_add( *cea, MSG_BRW_LAST_CHILD, avp ) );
		}
		
		/* The Relay application */
		if (g_pconf->disable_relay == 0) {
			dict_object_t * apprelay = NULL;
			dict_application_data_t relay;
			
			/* Find the Relay application in the dictionary */
			CHECK_FCT(  dict_search( DICT_APPLICATION, APPLICATION_BY_NAME, "Relay", &apprelay )  );
			CHECK_FCT(  dict_getval( apprelay, &relay )  );
			
			/* Find the model for this AVP */
			CHECK_FCT(  dict_search( DICT_AVP, AVP_BY_NAME, "Auth-Application-Id", &dict )  );

			/* Create the AVP */
			CHECK_FCT( msg_avp_new( dict, 0, &avp ) );

			/* Set its value */
			memset(&val, 0, sizeof(val));
			val.u32 = relay.application_id;
			CHECK_FCT( msg_avp_setvalue( avp, &val ) );

			/* Add it to the message */
			CHECK_FCT( msg_avp_add( *cea, MSG_BRW_LAST_CHILD, avp ) );
		}
	}
	
	/* Set the CEA header: end-to-end id */
	{
		msg_data_t * cea_data = NULL;
		
		CHECK_FCT(  msg_data( *cea, &cea_data )  );
		
		cea_data->msg_eteid = eteid;
		cea_data->msg_hbhid = hbhid;
	}
	
	return 0;
}
	
/* Stores the values of the CEA AVPs in the peer structure */
int _peer_cea_parse( msg_t * cea, _peer_t * peer ) 
{
	int i;
	msg_avp_t * avp = NULL;
	uint32_t errorcode = 0;
	uint32_t secidset = 0;
	
	CHECK_PARAMS(  VALIDATE_PEER( peer ) && cea  );

	CHECK_FCT(  msg_browse(cea, MSG_BRW_FIRST_CHILD, &avp, NULL)  );

	/* Now loop on all AVPs -- we will break when last AVP was parsed */
	while (avp) {
		msg_avp_data_t * avpdata;

		CHECK_FCT(  msg_avp_data( avp, &avpdata )  );

		if (avpdata->avp_flags & AVP_FLAG_VENDOR) {
			/* Ignore all vendor-specific AVPs in CEA ... because we don't support any currently */
			TRACE_DEBUG(FULL, "Ignored a vendor AVP in CEA");
			msg_dump_one(FULL, avp);
			goto next;
		}

		switch (avpdata->avp_code) {
			case AC_ORIGIN_HOST: /* Origin-Host */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CEA");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}

				/* Check that the remote peer is the peer we wanted to connect to */
				if (strncasecmp(peer->p_diamid, (char *)avpdata->avp_data->os.data, avpdata->avp_data->os.len)) {
					log_error("The CEA received from peer '%s' contains invalid Origin-Host (bad IP?), resetting connection", peer->p_diamid);
					msg_dump_one(0, avp);
					return EBADMSG;
				}
				break;

			case AC_ORIGIN_REALM: /* Origin-Realm */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CEA");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}

				/* Copy the value into peer->p_realm, as a lowercase string */
				free(peer->p_realm);
				CHECK_MALLOC(  peer->p_realm = (char *) malloc( avpdata->avp_data->os.len + 1 )  );
				for (i=0; i<avpdata->avp_data->os.len; i++) {
					peer->p_realm[i] = (char)tolower((char)avpdata->avp_data->os.data[i]);
				}
				peer->p_realm[i] = '\0';
				break;

			case AC_HOST_IP_ADDRESS: /* Host-IP-Address */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CEA");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}

				if (peer->p_peeraddr_sz == 0) {
					CHECK_MALLOC(  peer->p_peeraddr = (sSS *) malloc( sizeof(sSS) ) );
				} else {
					CHECK_MALLOC(  peer->p_peeraddr = (sSS *) realloc( peer->p_peeraddr, 
										(peer->p_peeraddr_sz + 1 ) * sizeof(sSS) )  );
				}

				/* Add this AVP as last element */
				CHECK_FCT( msg_avp_value_interpret( avp, &(peer->p_peeraddr[peer->p_peeraddr_sz]) ) );

				/* Update the number of elements in the array */
				peer->p_peeraddr_sz += 1;

				break;


			case AC_VENDOR_ID: /* Vendor-Id */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CEA");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}

				peer->p_vendor = avpdata->avp_data->u32;
				break;

			case AC_PRODUCT_NAME: /* Product-Name */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CEA");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}

				free(peer->p_prodname);

				CHECK_MALLOC(  peer->p_prodname = (char *) malloc( avpdata->avp_data->os.len + 1 )  );
				memcpy(peer->p_prodname, avpdata->avp_data->os.data, avpdata->avp_data->os.len);
				peer->p_prodname[avpdata->avp_data->os.len] = '\0';
				break;

			case AC_ORIGIN_STATE_ID: /* Origin-State-Id */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CEA");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}

				peer->p_orstate = avpdata->avp_data->u32;
				break;

			case AC_SUPPORTED_VENDOR_ID: /* Supported-Vendor-Id */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CEA");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}

				TRACE_DEBUG(FULL, "Info: The remote peer claims support for (a subset of) vendor %d features...", avpdata->avp_data->u32);
				break;

			case AC_VENDOR_SPECIFIC_APPLICATION_ID: /* Vendor-Specific-Application-Id (grouped)*/
				{
					peer_appl_t appdata;
					msg_avp_t * inavp = NULL;

					memset(&appdata, 0, sizeof(appdata));

					/* get the first child AVP */
					CHECK_FCT(  msg_browse(avp, MSG_BRW_FIRST_CHILD, &inavp, NULL)  );

					while (inavp) {
						msg_avp_data_t * inavpdata;
						CHECK_FCT(  msg_avp_data( inavp, &inavpdata )  );

						if (inavpdata->avp_flags & AVP_FLAG_VENDOR) {
							/* Ignore all vendor-specific AVPs in CER ... because we don't support any currently */
							TRACE_DEBUG(FULL, "Ignored a vendor AVP in Vendor-Specific-Application-Id AVP");
							msg_dump_one(FULL, avp);
							goto innext;
						}

						if (inavpdata->avp_data == NULL) {
							/* Ignore if the data is not set -- this is more like a sanity check */
							TRACE_DEBUG(FULL, "Ignored an AVP with unset value in Vendor-Specific-Application-Id AVP");
							msg_dump_one(FULL, avp);
							ASSERT(0); /* To check if this really happens, and understand why... */
							goto innext;
						}
						switch (inavpdata->avp_code) {
							case AC_VENDOR_ID: /* Vendor-Id */
								appdata.v = inavpdata->avp_data->u32;
								break;
							case AC_AUTH_APPLICATION_ID: /* Auth-Application-Id */
								appdata.a = inavpdata->avp_data->u32;
								appdata.t = APP_TYPE_AUTH;
								break;
							case AC_ACCT_APPLICATION_ID: /* Acct-Application-Id */
								appdata.a = inavpdata->avp_data->u32;
								appdata.t = APP_TYPE_ACCT;
								break;
							default: /* Other AVP */
								TRACE_DEBUG(FULL, "Ignored AVP in Vendor-Specific-Application-Id AVP");
								msg_dump_one(FULL, inavp);
						}

					innext:			
						/* Go to next in AVP */
						CHECK_FCT( msg_browse(inavp, MSG_BRW_NEXT, &inavp, NULL) );
					}

					CHECK_FCT(  _peer_struct_appl_add(peer, appdata.a, appdata.v, appdata.t)  );
				}
				break;

			case AC_AUTH_APPLICATION_ID: /* Auth-Application-Id */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CEA");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}

				CHECK_FCT(  _peer_struct_appl_add(peer, avpdata->avp_data->u32, 0, APP_TYPE_AUTH)  );
				break;

			case AC_ACCT_APPLICATION_ID: /* Acct-Application-Id */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CEA");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}

				CHECK_FCT(  _peer_struct_appl_add(peer, avpdata->avp_data->u32, 0, APP_TYPE_ACCT)  );
				break;

			case AC_INBAND_SECURITY_ID: /* Inband-Security-Id */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CEA");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}

				if (secidset) {
					log_error("Multiple Inband-Security-Id AVP in CEA from '%s', not supported.", peer->p_diamid);
					return EBADMSG;
				}
				
				/* Match the received ID with the list of sent proposals */
				while ( ! IS_LIST_EMPTY(&peer->p_sec_list) ) {
					_sec_item_t * si = (_sec_item_t *)(peer->p_sec_list.next);
					
					uti_list_unlink(_LIST(si));
					
					if ( ((sec_module_t *)(si->chain.o))->sec_insecid == avpdata->avp_data->u32 ) {
						/* We found it */
						peer->p_sec_hdl = si->hdl;
					} else {
						CHECK_FCT_DO( sec_modunlink(si->hdl), /* continue */ );
					}
					
					free(si);
				}
				
				secidset = 1;
				break;

			case AC_FIRMWARE_REVISION: /* Firmware-Revision */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CEA");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}

				peer->p_firmrev = avpdata->avp_data->u32;
				break;
				
			case AC_RESULT_CODE: /* Result-Code */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CEA");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}

				errorcode = avpdata->avp_data->u32;
				break;
				
			case AC_ERROR_MESSAGE: /* Error-Message */
				if (avpdata->avp_data == NULL) {
					/* Ignore if the data is not set -- this is more like a sanity check */
					TRACE_DEBUG(FULL, "Ignored an AVP with unset value in CEA");
					msg_dump_one(FULL, avp);
					ASSERT(0); /* To check if this really happens, and understand why... */
					goto next;
				}

				log_normal("The CEA received from peer '%s' contains an Error-Message AVP", peer->p_diamid);
				msg_dump_one(INFO, avp);
				break;

			default: /* Other AVP */
				TRACE_DEBUG(FULL, "Ignored AVP in CEA");
				msg_dump_one(FULL, avp);
		}

next:			
		/* Go to next AVP */
		CHECK_FCT( msg_browse(avp, MSG_BRW_NEXT, &avp, NULL) );
	}
	
	/* Ok, now check the error code */
	if ( errorcode != ER_DIAMETER_SUCCESS ) {
		/* We do not allow another error code here */
		log_error("Received CEA with Result-Code value %u from %s.\n", errorcode, peer->p_diamid);
		return EBADMSG;
	}
	
	return 0;
}
